"""Amcrest system module."""
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from datetime import datetime
from typing import Optional, Tuple

from .http import Http
from .utils import date_to_str, pretty, str_to_date


class System(Http):
    """Amcrest system class."""

    @property
    def current_time(self) -> datetime:
        ret = self.command("global.cgi?action=getCurrentTime")
        date_str = pretty(ret.content.decode())
        return str_to_date(date_str)

    @current_time.setter
    def current_time(self, date_value: datetime) -> bool:
        """
        According with API:
            The time format is "Y-M-D H-m-S". It is not be effected by Locales.
            TimeFormat in SetLocalesConfig

        Params:
            date = "Y-M-D H-m-S"
            Example: 2016-10-28 13:48:00

        Return: True
        """
        date_str = date_to_str(date_value)
        ret = self.command(f"global.cgi?action=setCurrentTime&time={date_str}")

        return "ok" in ret.content.decode().lower()

    @property
    async def async_current_time(self) -> datetime:
        ret = await self.async_command("global.cgi?action=getCurrentTime")
        date_str = pretty(ret.content.decode())
        return str_to_date(date_str)

    async def async_set_current_time(self, date_value: datetime) -> bool:
        """
        According with API:
            The time format is "Y-M-D H-m-S". It is not be effected by Locales.
            TimeFormat in SetLocalesConfig

        Params:
            date = "Y-M-D H-m-S"
            Example: 2016-10-28 13:48:00

        Return: True
        """
        date_str = date_to_str(date_value)
        ret = await self.async_command(
            f"global.cgi?action=setCurrentTime&time={date_str}"
        )

        return "ok" in ret.content.decode().lower()

    @property
    def general_config(self) -> str:
        return self._get_config("General")

    @property
    async def async_general_config(self) -> str:
        return await self._async_get_config("General")

    @property
    def version_http_api(self) -> str:
        ret = self.command("IntervideoManager.cgi?action=getVersion&Name=CGI")
        return ret.content.decode()

    @property
    async def async_version_http_api(self) -> str:
        ret = await self.async_command(
            "IntervideoManager.cgi?action=getVersion&Name=CGI"
        )
        return ret.content.decode()

    @property
    def software_information(self) -> Tuple[str, str]:
        sw_info = self._magic_box("getSoftwareVersion")
        return self._parse_sw_information(sw_info)

    @property
    async def async_software_information(self) -> Tuple[str, str]:
        sw_info = await self._async_magic_box("getSoftwareVersion")
        return self._parse_sw_information(sw_info)

    @staticmethod
    def _parse_sw_information(swinfo: str) -> Tuple[str, str]:
        if "," in swinfo:
            version, build_date = swinfo.split(",")
        else:
            version, build_date = swinfo.split()
        _, _, version = version.rpartition("=")
        _, _, build_date = build_date.rpartition(":")
        return version, build_date

    @property
    def hardware_version(self) -> str:
        return pretty(self._magic_box("getHardwareVersion"))

    @property
    async def async_hardware_version(self) -> str:
        return pretty(await self._async_magic_box("getHardwareVersion"))

    @property
    def device_type(self) -> str:
        return pretty(self._magic_box("getDeviceType"))

    @property
    async def async_device_type(self) -> str:
        return pretty(await self._async_magic_box("getDeviceType"))

    @property
    def serial_number(self) -> str:
        return pretty(self._magic_box("getSerialNo"))

    @property
    async def async_serial_number(self) -> str:
        return pretty(await self._async_magic_box("getSerialNo"))

    @property
    def machine_name(self) -> str:
        return pretty(self._magic_box("getMachineName"))

    @property
    async def async_machine_name(self) -> str:
        return pretty(await self._async_magic_box("getMachineName"))

    @property
    def system_information(self) -> str:
        """System information

        Including serial number, device type, processor, and serial.
        """
        return self._magic_box("getSystemInfo")

    @property
    async def async_system_information(self) -> str:
        return await self._async_magic_box("getSystemInfo")

    @property
    def vendor_information(self) -> str:
        return pretty(self._magic_box("getVendor"))

    @property
    async def async_vendor_information(self) -> str:
        return pretty(await self._async_magic_box("getVendor"))

    @property
    def onvif_information(self) -> str:
        ret = self.command(
            "IntervideoManager.cgi?action=getVersion&Name=Onvif"
        )
        return ret.content.decode().strip()

    @property
    async def async_onvif_information(self) -> str:
        ret = await self.async_command(
            "IntervideoManager.cgi?action=getVersion&Name=Onvif"
        )
        return ret.content.decode().strip()

    def config_backup(self, filename: Optional[str] = None) -> Optional[str]:
        ret = self.command("Config.backup?action=All")

        if not ret:
            return None

        if filename:
            with open(filename, "w+") as cfg:
                cfg.write(ret.content.decode())
            return None

        return ret.content.decode()

    async def async_config_backup(
        self, filename: Optional[str] = None
    ) -> Optional[str]:
        ret = await self.async_command("Config.backup?action=All")

        if not ret:
            return None

        if filename:
            with open(filename, "w+") as cfg:
                cfg.write(ret.content.decode())
            return None

        return ret.content.decode()

    @property
    def device_class(self) -> str:
        """
        During the development, device IP2M-841B didn't
        responde for this call, adding it anyway.
        """
        return pretty(self._magic_box("getDeviceClass"))

    @property
    async def async_device_class(self) -> str:
        return pretty(await self._async_magic_box("getDeviceClass"))

    def shutdown(self) -> str:
        """
        From the testings, shutdown acts like "reboot now"
        """
        return self._magic_box("shutdown")

    async def async_shutdown(self) -> str:
        return await self._async_magic_box("shutdown")

    def reboot(self, delay: Optional[int] = None) -> str:
        cmd = "reboot"
        if delay:
            cmd += f"&delay={delay}"
        return self._magic_box(cmd)

    async def async_reboot(self, delay: Optional[int] = None) -> str:
        cmd = "reboot"
        if delay:
            cmd += f"&delay={delay}"
        return await self._async_magic_box(cmd)

    def onvif_login_check(self, setCheck: bool = False) -> str:
        """
        Allows the other non-admin accounts to use ONVIF.
        Currently only the 'admin' account can use ONVIF.
        """
        cmd = "configManager.cgi?action=setConfig"
        cmd += "&UserGlobal.OnvifLoginCheck={0}".format(str(setCheck).lower())
        ret = self.command(cmd)

        return ret.content.decode()

    async def async_onvif_login_check(self, setCheck: bool = False) -> str:
        """
        Allows the other non-admin accounts to use ONVIF.
        Currently only the 'admin' account can use ONVIF.
        """
        cmd = "configManager.cgi?action=setConfig"
        cmd += "&UserGlobal.OnvifLoginCheck={0}".format(str(setCheck).lower())
        ret = await self.async_command(cmd)

        return ret.content.decode()



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import time
from amcrest import AmcrestCamera

amcrest = AmcrestCamera('192.168.1.6', 80, 'admin', 'super-password')

amcrest.camera.zoom_out("start")
time.sleep(0.5)
amcrest.camera.zoom_out("stop")

# Snapshot
amcrest.camera.snapshot(path_file="/tmp/snapshot.jpg")



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

import datetime
from amcrest import AmcrestCamera

amcrest = AmcrestCamera("192.168.1.10", 80, "admin", "super_password")
camera = amcrest.camera

end_time = datetime.datetime.now()
start_time = end_time - datetime.timedelta(hours=1)

for text in camera.find_files(start_time, end_time):
    for line in text.split("\r\n"):
        key, value = list(line.split("=", 1) + [None])[:2]
        if key.endswith(".FilePath"):
            print("Found file {}".format(value))
            if value.endswith(".jpg"):
                file_name = value

if file_name:
    print("Downloading {}...".format(file_name))
    with open("snapshot.jpg", "w") as file:
        file.write(camera.download_file(file_name))
    print("Saved as {}!".format("snapshot.jpg"))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from datetime import datetime
from typing import AsyncIterator, Iterator

from .http import Http
from .utils import date_to_str


class Log(Http):
    def log_clear_all(self) -> str:
        ret = self.command("log.cgi?action=clear")
        return ret.content.decode()

    async def async_log_clear_all(self) -> str:
        ret = await self.async_command("log.cgi?action=clear")
        return ret.content.decode()

    def log_show(self, start_time: datetime, end_time: datetime) -> str:
        start = date_to_str(start_time)
        end = date_to_str(end_time)
        ret = self.command(
            "Log.backup?action=All&"
            f"condition.StartTime={start}&condition.EndTime={end}"
        )
        return ret.content.decode()

    async def async_log_show(
        self, start_time: datetime, end_time: datetime
    ) -> str:
        start = date_to_str(start_time)
        end = date_to_str(end_time)
        ret = await self.async_command(
            "Log.backup?action=All&"
            f"condition.StartTime={start}&condition.EndTime={end}"
        )
        return ret.content.decode()

    def log_find_start(self, start_time: datetime, end_time: datetime) -> str:
        start = date_to_str(start_time)
        end = date_to_str(end_time)
        ret = self.command(
            "log.cgi?action=startFind&"
            f"condition.StartTime={start}&condition.EndTime={end}"
        )

        return ret.content.decode()

    async def async_log_find_start(
        self, start_time: datetime, end_time: datetime
    ) -> str:
        start = date_to_str(start_time)
        end = date_to_str(end_time)
        ret = await self.async_command(
            "log.cgi?action=startFind&"
            f"condition.StartTime={start}&condition.EndTime={end}"
        )

        return ret.content.decode()

    def log_find_next(self, token: str, count: int = 100) -> str:
        ret = self.command(
            f"log.cgi?action=doFind&token={token}&count={count}"
        )
        return ret.content.decode()

    async def async_log_find_next(self, token: str, count: int = 100) -> str:
        ret = await self.async_command(
            f"log.cgi?action=doFind&token={token}&count={count}"
        )
        return ret.content.decode()

    def log_find_stop(self, token: str) -> str:
        ret = self.command(f"log.cgi?action=stopFind&token={token}")
        return ret.content.decode()

    async def async_log_find_stop(self, token: str) -> str:
        ret = await self.async_command(
            f"log.cgi?action=stopFind&token={token}"
        )
        return ret.content.decode()

    def log_find(
        self, start_time: datetime, end_time: datetime
    ) -> Iterator[str]:
        token = self.log_find_start(start_time, end_time).strip().split("=")[1]

        while True:
            content = self.log_find_next(token)
            tag, _, count = content.split("\r\n", 1)[0].partition("=")

            yield content

            if tag != "found" or int(count) == 0:
                break

        self.log_find_stop(token)

    async def async_log_find(
        self, start_time: datetime, end_time: datetime
    ) -> AsyncIterator[str]:
        token = (
            (await self.async_log_find_start(start_time, end_time))
            .strip()
            .split("=")[1]
        )

        while True:
            content = await self.async_log_find_next(token)
            tag, _, count = content.split("\r\n", 1)[0].partition("=")

            yield content

            if tag != "found" or int(count) == 0:
                break

        await self.async_log_find_stop(token)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import logging
import shutil
from typing import Optional, Union, overload
from typing_extensions import Literal

from urllib3.exceptions import HTTPError
from urllib3.response import HTTPResponse

from .exceptions import CommError
from .http import Http, TimeoutT

_LOGGER = logging.getLogger(__name__)


class Snapshot(Http):
    @property
    def snapshot_config(self) -> str:
        return self._get_config("Snap")

    @property
    async def async_snapshot_config(self) -> str:
        return await self._async_get_config("Snap")

    @overload
    def snapshot(
        self,
        *,
        channel: Optional[int] = ...,
        path_file: Optional[str] = ...,
        timeout: TimeoutT = ...,
        stream: Literal[True] = ...,
    ) -> HTTPResponse:
        ...

    @overload
    def snapshot(
        self,
        *,
        channel: Optional[int] = ...,
        path_file: Optional[str] = ...,
        timeout: TimeoutT = ...,
        stream: Literal[False] = ...,
    ) -> bytes:
        ...

    def snapshot(
        self,
        *,
        channel: Optional[int] = None,
        path_file: Optional[str] = None,
        timeout: TimeoutT = None,
        stream: bool = True,
    ) -> Union[bytes, HTTPResponse]:
        """
        Args:

            channel:
                Video input channel number

                If no channel param is used, don't send channel parameter
                so camera will use its default channel

            path_file:
                If path_file is provided, save the snapshot
                in the path

        Return:
            raw from http request if stream is True
            response content if stream is False
        """
        cmd = "snapshot.cgi"
        if channel is not None:
            cmd += f"?channel={channel}"
        ret = self.command(cmd, timeout_cmd=timeout, stream=stream)

        if path_file:
            with open(path_file, "wb") as out_file:
                if stream:
                    try:
                        shutil.copyfileobj(ret.raw, out_file)
                    except HTTPError as error:
                        _LOGGER.debug(
                            "%s Snapshot to file failed due to error: %s",
                            self,
                            repr(error),
                        )
                        raise CommError(error) from error
                else:
                    out_file.write(ret.content)

        return ret.raw if stream else ret.content

    async def async_snapshot(
        self, *, channel: Optional[int] = None, timeout: TimeoutT = None
    ) -> bytes:
        cmd = "snapshot.cgi"
        if channel is not None:
            cmd += f"?channel={channel}"
        ret = await self.async_command(cmd, timeout_cmd=timeout)

        return ret.content



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

import re
from datetime import datetime

# pylint: disable=no-name-in-module
from distutils import util
from typing import List, Tuple, Union

DATEFMT = "%Y-%m-%d %H:%M:%S"
PRECISION = 2


def clean_url(url: str) -> str:
    host = re.sub(r"^http[s]?://", "", url, flags=re.IGNORECASE)
    host = re.sub(r"/$", "", host)
    return host


def pretty(value: str, delimiter: str = "=") -> str:
    """Format string key=value."""
    return value.strip().rpartition(delimiter)[2]


def date_to_str(value: datetime) -> str:
    return value.strftime(DATEFMT)


def str_to_date(value: str) -> datetime:
    return datetime.strptime(value, DATEFMT)


def percent(part: float, whole: float) -> str:
    """Convert data to percent"""
    return str(round(100 * part / whole, PRECISION))


def str2bool(value: Union[str, int]) -> bool:
    """
    Args:
        value - text to be converted to boolean
         True values: y, yes, true, t, on, 1
         False values: n, no, false, off, 0
    """
    if isinstance(value, str):
        return bool(util.strtobool(value))
    return bool(value)


def to_unit(
    value: Union[None, str, int, float], unit: str = "B"
) -> Tuple[str, str]:
    """Convert bytes to give unit."""
    byte_array = ["B", "KB", "MB", "GB", "TB"]

    if unit not in byte_array:
        raise ValueError(f"Unit {unit} missing from known units")

    if value is None:
        return "unknown", unit
    if isinstance(value, (int, float)):
        value_f = value
    else:
        try:
            value_f = float(value)
        except (TypeError, ValueError):
            return "unknown", unit

    result = value_f / 1024 ** byte_array.index(unit)
    return str(round(result, PRECISION)), unit


def extract_audio_video_enabled(param: str, resp: str) -> List[bool]:
    """Extract if any audio/video stream enabled from response."""
    parts = [
        part.rpartition("=")[2] == "true"
        for part in resp.split()
        if f".{param}Enable=" in part
    ]
    return parts


def enable_audio_video_cmd(param: str, enable: bool, channel: int) -> str:
    """Return command to enable/disable all audio/video streams."""
    formats = [("Extra", 3), ("Main", 4)]
    if param == "Video":
        formats.append(("Snap", 3))

    set_enable = str(enable).lower()
    cmds = ["configManager.cgi?action=setConfig"]
    cmds.extend(
        f"Encode[{channel}].{fmt}Format[{i}].{param}Enable={set_enable}"
        for fmt, num in formats
        for i in range(num)
    )
    return "&".join(cmds)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from .http import Http


# This API is based on an undocumented API of Amcrest Cameras.
# This API may change without notice
class PrivacyMode(Http):
    def set_privacy(self, mode: bool) -> str:
        """
        Params:
            mode         - True to enable privacy mode, false to disable

        Turns on privacy mode on or off in the camera
        """
        ret = self.command(
            f"configManager.cgi?action=setConfig&LeLensMask[0].Enable={mode}"
        )
        return ret.content.decode()

    async def async_set_privacy(self, mode: bool) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=setConfig&LeLensMask[0].Enable={mode}"
        )
        return ret.content.decode()

    def privacy_config(self) -> str:
        return self._get_config("LeLensMask")

    async def async_privacy_config(self) -> str:
        return await self._async_get_config("LeLensMask")



"""Amcrest NAS."""
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from .http import Http


class Nas(Http):
    """Amcrest methods to handle NAS."""

    @property
    def nas_information(self) -> str:
        """Return NAS information."""
        return self._get_config("NAS")

    @property
    async def async_nas_information(self) -> str:
        """Return NAS information."""
        return await self._async_get_config("NAS")



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import logging
import shutil
from typing import Optional

from urllib3.exceptions import HTTPError

from .exceptions import CommError
from .http import Http

_LOGGER = logging.getLogger(__name__)


class Special(Http):
    def realtime_stream(
        self,
        *,
        channel: int = 1,
        typeno: int = 0,
        path_file: Optional[str] = None,
    ) -> bytes:
        """
        If the stream is redirect to a file, use mplayer tool to
        visualize the video record

        camera.realtime_stream(path_file="/home/user/Desktop/myvideo)
        $ mplayer /home/user/Desktop/myvideo
        """
        ret = self.command(
            f"realmonitor.cgi?action=getStream&channel={channel}&"
            f"subtype={typeno}",
            stream=True,
        )

        if path_file:
            try:
                with open(path_file, "wb") as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug(
                    "%s Realtime stream capture to file failed due "
                    "to error: %s",
                    self,
                    repr(error),
                )
                raise CommError(error) from error

        return ret.raw

    def rtsp_url(self, *, channel: int = 1, typeno: int = 0) -> str:
        """
        Return RTSP streaming url

        Params:
            channel: integer, the video channel index which starts from 1,
                     default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        return self._build_rtsp_url(self.rtsp_config, channel, typeno)  # type: ignore[attr-defined]

    async def async_rtsp_url(
        self, *, channel: int = 1, typeno: int = 0
    ) -> str:
        rtsp_config = await self.async_rtsp_config  # type: ignore[attr-defined]
        return self._build_rtsp_url(rtsp_config, channel, typeno)

    def _build_rtsp_url(
        self, rtsp_config: str, channel: int, typeno: int
    ) -> str:
        cmd = f"cam/realmonitor?channel={channel}&subtype={typeno}"

        try:
            port_num = [
                x.split("=")[1]
                for x in rtsp_config.split()
                if x.startswith("table.RTSP.Port=")
            ][0]
        except IndexError:
            port = ""
        else:
            port = ":{}".format(port_num)

        return f"rtsp://{self._user}:{self._password}@{self._host}{port}/{cmd}"

    # pylint: disable=pointless-string-statement
    """
    11/05/2016

    Looks like the mjpeg support was removed from the mainstream in one of
    the firmware updates, so subtype=0 is not working.

    I have tested with the below firmware and still not working
    in the mainstream:
        Amcrest_IPC-AWXX_Eng_N_V2.420.AC00.15.R.20160908.bin

    Based on above, the cmd we send to camera should use: subtype=1 until
    mainstream is back.

    Simple translation:
        subtype=0 for mainstream
        subtype=1 for substream

    Tested on device: IP2M-841B

    Users complaining here too:
    https://amcrest.com/forum/technical-discussion-f3/lost-mjpeg-encode-for-main-stream-after-firmware-u-t1516.html
    """

    def mjpeg_url(self, *, channel: int = 1, typeno: int = 0) -> str:
        """
        Return MJPEG streaming url

        Params:
            channel: integer, the video channel index which starts from 1,
                     default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        return (
            f"{self._base_url}mjpg/video.cgi?"
            f"channel={channel}&subtype={typeno}"
        )

    def mjpg_stream(
        self,
        *,
        channel: int = 1,
        typeno: int = 0,
        path_file: Optional[str] = None,
    ) -> bytes:
        """
        Params:
            channel: integer, the video channel index which starts from 1,
                     default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        cmd = self.mjpeg_url(channel=channel, typeno=typeno)
        ret = self.command(cmd, stream=True)

        if path_file:
            try:
                with open(path_file, "wb") as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug(
                    "%s MJPEG stream capture to file failed due to error: %s",
                    self,
                    repr(error),
                )
                raise CommError(error) from error

        return ret.raw



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from amcrest.exceptions import CommError
from amcrest.http import Http
from amcrest.utils import pretty


class Record(Http):
    @property
    def record_capability(self) -> str:
        ret = self.command("recordManager.cgi?action=getCaps")
        return ret.content.decode()

    async def async_record_capability(self) -> str:
        ret = await self.async_command("recordManager.cgi?action=getCaps")
        return ret.content.decode()

    @property
    def record_config(self) -> str:
        return self._get_config("Record")

    @record_config.setter
    def record_config(self, rec_opt: str) -> str:
        """
        rec_opt is the Record options listed as example below:

        +----------------------+--------+------------------------------------+
        | ParamName            | Value  |   Description                      |
        +----------------------+--------+------------------------------------+
        | Record[ch].PreRecord |Integer | Range is [0-300]                   |
        |                      |        | Prerecord seconds, 0 no prerecord  |
        |                      |        | ch (Channel number) starts from 0  |
        +----------------------|--------|------------------------------------+
        | Record[ch].          |        | wd (week day)                      |
        | TimeSection[wd][ts]  | string | range is [0-6] (Sun/Sat)           |
        |                      |        |                                    |
        |                      |        | ts (time section) range is [0-23]  |
        |                      |        | time section table index           |
        |                      |        |                                    |
        |                      |        | Format: mas hh:mm:ss-hh:mm:ss      |
        |                      |        | Mask: [0-65535], hh: [0-24],       |
        |                      |        | mm: [0-59], ss: [0-59]             |
        |                      |        | Mask indicate record type by bits: |
        |                      |        | Bit0: regular record               |
        |                      |        | Bit1: motion detection record      |
        |                      |        | Bit2: alarm record                 |
        |                      |        | Bit3: card record                  |
        +----------------------+--------+------------------------------------+

        Example:
        Record[0].TimeSection[0][0]=6 00:00:00-23:59:59


        rec_opt format:
        <paramName>=<paramValue>[&<paramName>=<paramValue>...]
        """

        ret = self.command(f"configManager.cgi?action=setConfig&{rec_opt}")
        return ret.content.decode()

    @property
    async def async_record_config(self) -> str:
        return await self._async_get_config("Record")

    async def async_set_record_config(self, rec_opt: str) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=setConfig&{rec_opt}"
        )
        return ret.content.decode()

    @property
    def media_global_config(self) -> str:
        return self._get_config("MediaGlobal")

    async def async_media_global_config(self) -> str:
        return await self._async_get_config("MediaGlobal")

    @property
    def record_mode(self) -> str:
        return self.get_record_mode()

    @record_mode.setter
    def record_mode(self, record_opt: int) -> None:
        self.set_record_mode(record_opt)

    @property
    async def async_record_mode(self) -> str:
        return await self.async_get_record_mode()

    def get_record_mode(self, channel: int = 0) -> str:
        # Certain compatible cameras, such Dahua doorbells, will return an
        # error code 400 when using this API.  Gracefully catch this and return
        # "Unknown" in this case.
        try:
            ret = self._get_config("RecordMode")
        except CommError:
            return "Unknown"
        return self._process_record_mode(ret, channel)

    async def async_get_record_mode(self, channel: int = 0) -> str:
        # Similar behavior as above
        try:
            ret = await self._async_get_config("RecordMode")
        except CommError:
            return "Unknown"
        return self._process_record_mode(ret, channel)

    def _process_record_mode(self, record_mode: str, channel: int) -> str:
        status_code = {0: "Automatic", 1: "Manual", 2: "Stop"}

        statuses = [
            pretty(s) for s in record_mode.split() if f"[{channel}].Mode=" in s
        ]
        if len(statuses) != 1:
            return "Unknown"

        status = int(statuses[0])
        if status not in status_code:
            return "Unknown"

        return status_code[status]

    def set_record_mode(self, record_opt: int, *, channel: int = 0) -> str:
        """
        Params:

        channel:
        video index, start from 0

        record_opt:
        +----------------------------+-----------------+-------------------+
        | ParamName                  | ParamValue type | Description       |
        +----------------------------+-----------------+-------------------+
        | RecordMode[channel].Mode   | integer         | Range os {0, 1, 2}|
        |                            |                 | 0: automatically  |
        |                            |                 | 1: manually       |
        |                            |                 | 2: stop record    |
        +----------------------------+-----------------+-------------------+

        record_opt format:
        <paramName>=<paramValue>[&<paramName>=<paramValue>...]
        """
        ret = self.command(
            "configManager.cgi?action=setConfig&RecordMode"
            f"[{channel}].Mode={record_opt}"
        )
        return ret.content.decode()

    async def async_set_record_mode(
        self, record_opt: int, *, channel: int = 0
    ) -> str:
        ret = await self.async_command(
            "configManager.cgi?action=setConfig&RecordMode"
            f"[{channel}].Mode={record_opt}"
        )
        return ret.content.decode()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from typing import List

from . import utils
from .http import Http


class Video(Http):
    @property
    def video_max_extra_streams(self) -> int:
        ret = self._magic_box("getProductDefinition&name=MaxExtraStream")
        return int(utils.pretty(ret))

    @property
    async def async_video_max_extra_streams(self) -> int:
        ret = await self._async_magic_box(
            "getProductDefinition&name=MaxExtraStream"
        )
        return int(utils.pretty(ret))

    @property
    def video_color_config(self) -> str:
        return self._get_config("VideoColor")

    @property
    async def async_video_color_config(self) -> str:
        return await self._async_get_config("VideoColor")

    @property
    def encode_capability(self) -> str:
        ret = self.command("encode.cgi?action=getCaps")
        return ret.content.decode()

    @property
    async def async_encode_capability(self) -> str:
        ret = await self.async_command("encode.cgi?action=getCaps")
        return ret.content.decode()

    def encode_config_capability(self, channel: int) -> str:
        ret = self.command(
            f"encode.cgi?action=getConfigCaps&channel={channel}"
        )
        return ret.content.decode()

    async def async_encode_config_capability(self, channel: int) -> str:
        ret = await self.async_command(
            f"encode.cgi?action=getConfigCaps&channel={channel}"
        )
        return ret.content.decode()

    @property
    def encode_media(self) -> str:
        return self._get_config("Encode")

    @property
    async def async_encode_media(self) -> str:
        return await self._async_get_config("Encode")

    @property
    def encode_region_interested(self) -> str:
        return self._get_config("VideoEncodeROI")

    @property
    async def async_encode_region_interested(self) -> str:
        return await self._async_get_config("VideoEncodeROI")

    @property
    def video_channel_title(self) -> str:
        return self._get_config("ChannelTitle")

    @property
    async def async_video_channel_title(self) -> str:
        return await self._async_get_config("ChannelTitle")

    # pylint: disable=invalid-name
    @property
    def video_input_channels_device_supported(self) -> str:
        ret = self.command("devVideoInput.cgi?action=getCollect")
        return ret.content.decode()

    @property
    async def async_video_input_channels_device_supported(self) -> str:
        ret = await self.async_command("devVideoInput.cgi?action=getCollect")
        return ret.content.decode()

    # pylint: disable=invalid-name
    @property
    def video_output_channels_device_supported(self) -> str:
        ret = self.command("devVideoOutput.cgi?action=getCollect")
        return ret.content.decode()

    @property
    async def async_video_output_channels_device_supported(self) -> str:
        ret = await self.async_command("devVideoOutput.cgi?action=getCollect")
        return ret.content.decode()

    # pylint: disable=invalid-name
    @property
    def video_max_remote_input_channels(self) -> str:
        return self._magic_box(
            "getProductDefinition&name=MaxRemoteInputChannels"
        )

    @property
    async def async_video_max_remote_input_channels(self) -> str:
        return await self._async_magic_box(
            "getProductDefinition&name=MaxRemoteInputChannels"
        )

    @property
    def video_standard(self) -> str:
        return self._get_config("VideoStandard")

    @video_standard.setter
    def video_standard(self, std: str) -> str:
        return self._get_config(f"setConfig&VideoStandard={std}")

    @property
    async def async_video_standard(self) -> str:
        return await self._async_get_config("VideoStandard")

    async def async_set_video_standard(self, std: str) -> str:
        return await self._async_get_config(f"setConfig&VideoStandard={std}")

    @property
    def video_widget_config(self) -> str:
        return self._get_config("VideoWidget")

    @property
    async def async_video_widget_config(self) -> str:
        return await self._async_get_config("VideoWidget")

    def video_input_capability(self, channel: int) -> str:
        ret = self.command(
            f"devVideoInput.cgi?action=getCaps&channel={channel}"
        )
        return ret.content.decode()

    async def async_video_input_capability(self, channel: int) -> str:
        ret = await self.async_command(
            f"devVideoInput.cgi?action=getCaps&channel={channel}"
        )
        return ret.content.decode()

    def video_coordinates_current_window(self, channel: int) -> str:
        ret = self.command(
            f"devVideoInput.cgi?action=getCurrentWindow&channel={channel}"
        )
        return ret.content.decode()

    async def async_video_coordinates_current_window(
        self, channel: int
    ) -> str:
        ret = await self.async_command(
            f"devVideoInput.cgi?action=getCurrentWindow&channel={channel}"
        )
        return ret.content.decode()

    @property
    def video_in_options(self) -> str:
        return self._get_config("VideoInOptions")

    @property
    async def async_video_in_options(self) -> str:
        return await self._async_get_config("VideoInOptions")

    def video_in_option(
        self, param: str, *, profile: str = "Day"
    ) -> List[str]:
        """
        Return video input option.

        Params:
            param - parameter, such as 'DayNightColor'
            profile - 'Day', 'Night' or 'Normal'
        """
        if profile == "Day":
            field = param
        else:
            field = f"{profile}Options.{param}"
        return [
            utils.pretty(opt)
            for opt in self.video_in_options.split()
            if f"].{field}=" in opt
        ]

    async def async_video_in_option(
        self, param: str, *, profile: str = "Day"
    ) -> List[str]:
        """
        Return video input option.

        Params:
            param - parameter, such as 'DayNightColor'
            profile - 'Day', 'Night' or 'Normal'
        """
        if profile == "Day":
            field = param
        else:
            field = f"{profile}Options.{param}"
        return [
            utils.pretty(opt)
            for opt in (await self.async_video_in_options).split()
            if f"].{field}=" in opt
        ]

    def set_video_in_option(
        self, param: str, value: str, *, profile: str = "Day", channel: int = 0
    ) -> str:
        if profile == "Day":
            field = param
        else:
            field = f"{profile}Options.{param}"
        ret = self.command(
            "configManager.cgi?action=setConfig"
            f"&VideoInOptions[{channel}].{field}={value}"
        )
        return ret.content.decode()

    async def async_set_video_in_option(
        self, param: str, value: str, *, profile: str = "Day", channel: int = 0
    ) -> str:
        if profile == "Day":
            field = param
        else:
            field = f"{profile}Options.{param}"
        ret = await self.async_command(
            "configManager.cgi?action=setConfig"
            f"&VideoInOptions[{channel}].{field}={value}"
        )
        return ret.content.decode()

    @property
    def day_night_color(self) -> int:
        """
        Return Day & Night Color Mode for Day profile.

        Result is 0: always multicolor
                  1: autoswitch along with brightness
                  2: always monochrome
        """
        return self.get_day_night_color(channel=0)

    @day_night_color.setter
    def day_night_color(self, value: int) -> str:
        return self.set_video_in_option("DayNightColor", str(value), channel=0)

    @property
    async def async_day_night_color(self) -> int:
        """Return Day & Night Color Mode for Day profile."""
        return await self.async_get_day_night_color(channel=0)

    def get_day_night_color(self, channel: int) -> int:
        """
        Return Day & Night Color Mode for Day profile.

        Result is 0: always multicolor
                  1: autoswitch along with brightness
                  2: always monochrome
        """
        values = [int(x) for x in self.video_in_option("DayNightColor")]
        return values[channel]

    async def async_get_day_night_color(self, channel: int) -> int:
        """
        Return Day & Night Color Mode for Day profile.

        Result is 0: always multicolor
                  1: autoswitch along with brightness
                  2: always monochrome
        """
        values = [
            int(x) for x in await self.async_video_in_option("DayNightColor")
        ]
        return values[channel]

    def set_day_night_color(self, value: int, channel: int) -> str:
        return self.set_video_in_option(
            "DayNightColor", str(value), channel=channel
        )

    async def async_set_day_night_color(self, value: int, channel: int) -> str:
        return await self.async_set_video_in_option(
            "DayNightColor", str(value), channel=channel
        )

    @property
    def smart_ir(self) -> bool:
        """Return if SmartIR is on."""
        return self.is_smart_ir_on(channel=0)

    @smart_ir.setter
    def smart_ir(self, value: bool) -> str:
        return self.set_smart_ir(value, channel=0)

    @property
    async def async_smart_ir(self) -> bool:
        """Return if SmartIR is on."""
        return await self.async_is_smart_ir_on(channel=0)

    def is_smart_ir_on(self, channel: int) -> bool:
        """Return if SmartIR is on."""
        return self.video_in_option("InfraRed")[channel] == "false"

    async def async_is_smart_ir_on(self, channel: int) -> bool:
        """Return if SmartIR is on."""
        return (await self.async_video_in_option("InfraRed"))[
            channel
        ] == "false"

    def set_smart_ir(self, value: bool, channel: int) -> str:
        # It's not clear why from the HTTP API SDK doc, but setting InfraRed
        # to false sets the Night Vision Mode to SmartIR, whereas setting it
        # to true sets the Night Vision Mode to OFF. Night Vision Mode has a
        # third setting of Manual, but that must be selected some other way
        # via the HTTP API.
        str_value = "false" if value else "true"
        return self.set_video_in_option("InfraRed", str_value, channel=channel)

    async def async_set_smart_ir(self, value: bool, channel: int) -> str:
        # It's not clear why from the HTTP API SDK doc, but setting InfraRed
        # to false sets the Night Vision Mode to SmartIR, whereas setting it
        # to true sets the Night Vision Mode to OFF. Night Vision Mode has a
        # third setting of Manual, but that must be selected some other way
        # via the HTTP API.
        str_value = "false" if value else "true"
        return await self.async_set_video_in_option(
            "InfraRed", str_value, channel=channel
        )

    @property
    def video_enabled(self) -> bool:
        """Return if main video stream enabled."""
        return self.is_video_enabled(channel=0)

    @video_enabled.setter
    def video_enabled(self, enable: bool) -> None:
        self.set_video_enabled(enable, channel=0)

    @property
    async def async_video_enabled(self) -> bool:
        """Return if main video stream enabled."""
        return await self.async_is_video_enabled(channel=0)

    def is_video_enabled(
        self, channel: int, *, stream: str = "Main", stream_type: int = 0
    ) -> bool:
        """Return if given video stream enabled.

        The stream should be either "Main", "Extra", or "Snap".  For the main
        stream, the stream type selects if it should read regular (0), motion
        detection (1), alarm (2), or emergency (3) encode settings.  The snap
        stream has the same settings for regular, motion detection, and alarm
        stream types.  For the extra stream, the stream type selects which
        stream should be read, and is zero indexed from 0 to 2 for streams 1 to
        3.
        """
        is_enabled = utils.extract_audio_video_enabled(
            f"{stream}Format[{stream_type}].Video", self.encode_media
        )
        return is_enabled[channel]

    async def async_is_video_enabled(
        self, channel: int, *, stream: str = "Main", stream_type: int = 0
    ) -> bool:
        """Return if given video stream enabled."""
        is_enabled = utils.extract_audio_video_enabled(
            f"{stream}Format[{stream_type}].Video",
            await self.async_encode_media,
        )
        return is_enabled[channel]

    def set_video_enabled(self, enable: bool, channel: int) -> None:
        self.command(utils.enable_audio_video_cmd("Video", enable, channel))
        self.set_smart_ir(enable, channel)

    async def async_set_video_enabled(
        self, enable: bool, channel: int
    ) -> None:
        await self.async_command(
            utils.enable_audio_video_cmd("Video", enable, channel)
        )
        await self.async_set_smart_ir(enable, channel)

    @property
    def video_out_options(self) -> str:
        return self._get_config("VideoOut")

    @property
    async def async_video_out_options(self) -> str:
        return await self._async_get_config("VideoOut")



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
from .audio import Audio
from .event import Event
from .exceptions import AmcrestError, CommError, LoginError  # noqa: F401
from .log import Log
from .media import Media
from .motion_detection import MotionDetection
from .nas import Nas
from .network import Network
from .ptz import Ptz
from .record import Record
from .snapshot import Snapshot
from .special import Special
from .storage import Storage
from .system import System
from .user_management import UserManagement
from .video import Video
from .privacy_mode import PrivacyMode


class AmcrestCamera:
    """Amcrest camera object implementation."""

    def __init__(
        self,
        host,
        port,
        user,
        password,
        *,
        verbose=True,
        protocol="http",
        ssl_verify=True,
        retries_connection=None,
        timeout_protocol=None,
    ) -> None:
        super().__init__()
        self.camera = ApiWrapper(
            host=host,
            port=port,
            user=user,
            password=password,
            verbose=verbose,
            protocol=protocol,
            ssl_verify=ssl_verify,
            retries_connection=retries_connection,
            timeout_protocol=timeout_protocol,
        )


# pylint: disable=too-many-ancestors
class ApiWrapper(
    System,
    Network,
    MotionDetection,
    Snapshot,
    UserManagement,
    Event,
    Audio,
    Record,
    Video,
    Log,
    Ptz,
    Special,
    Storage,
    Nas,
    Media,
    PrivacyMode,
):
    pass



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import logging
import re
from typing import (
    Any,
    AsyncIterator,
    AsyncIterable,
    Dict,
    Iterator,
    Iterable,
    List,
    Optional,
    Tuple,
)

from requests import RequestException
from urllib3.exceptions import HTTPError

from .exceptions import CommError
from .http import Http, TimeoutT
from .utils import pretty

_LOGGER = logging.getLogger(__name__)
_REG_PARSE_KEY_VALUE = re.compile(r"(?P<key>.+?)(?:=)(?P<value>.+?)(?:;|$)")
_REG_PARSE_MALFORMED_JSON = re.compile(
    r'(?P<key>"[^"\\]*(?:\\.[^"\\]*)*"|[^\s"]+)\s:\s(?P<value>"[^"\\]*(?:\\.[^"\\]*)*"|[^\s"]+)'
)


def _event_lines(ret: Iterable[str]) -> Iterator[str]:
    line = []
    for char in ret:
        line.append(char)
        if line[-2:] == ["\r", "\n"]:
            yield "".join(line).strip()
            line.clear()


async def _async_event_lines(ret: AsyncIterable[str]) -> AsyncIterator[str]:
    line = []
    async for char in ret:
        line.append(char)
        if line[-2:] == ["\r", "\n"]:
            yield "".join(line).strip()
            line.clear()


class Event(Http):
    def event_handler_config(self, handlername: str) -> str:
        return self._get_config(handlername)

    async def async_event_handler_config(self, handlername: str) -> str:
        return await self._async_get_config(handlername)

    def alarm_handler(self, alarm_name: str) -> str:
        ret = self.command("alarm.cgi?action={alarm_name}")
        return ret.content.decode()

    async def async_alarm_handler(self, alarm_name: str) -> str:
        ret = await self.async_command("alarm.cgi?action={alarm_name}")
        return ret.content.decode()

    @property
    def alarm_config(self) -> str:
        return self.event_handler_config("Alarm")

    @property
    async def async_alarm_config(self) -> str:
        return await self.async_event_handler_config("Alarm")

    @property
    def alarm_out_config(self) -> str:
        return self.event_handler_config("AlarmOut")

    @property
    async def async_alarm_out_config(self) -> str:
        return await self.async_event_handler_config("AlarmOut")

    @property
    def alarm_input_channels(self) -> str:
        return self.alarm_handler("getInSlots")

    @property
    async def async_alarm_input_channels(self) -> str:
        return await self.async_alarm_handler("getInSlots")

    @property
    def alarm_output_channels(self) -> str:
        return self.alarm_handler("getOutSlots")

    @property
    async def async_alarm_output_channels(self) -> str:
        return await self.async_alarm_handler("getOutSlots")

    @property
    def alarm_states_input_channels(self) -> str:
        return self.alarm_handler("getInState")

    @property
    async def async_alarm_states_input_channels(self) -> str:
        return await self.async_alarm_handler("getInState")

    @property
    def alarm_states_output_channels(self) -> str:
        return self.alarm_handler("getOutState")

    @property
    async def async_alarm_states_output_channels(self) -> str:
        return await self.async_alarm_handler("getOutState")

    @property
    def video_blind_detect_config(self) -> str:
        return self.event_handler_config("BlindDetect")

    @property
    async def async_video_blind_detect_config(self) -> str:
        return await self.async_event_handler_config("BlindDetect")

    @property
    def video_loss_detect_config(self) -> str:
        return self.event_handler_config("LossDetect")

    @property
    async def async_video_loss_detect_config(self) -> str:
        return await self.async_event_handler_config("LossDetect")

    @property
    def event_login_failure(self) -> str:
        return self.event_handler_config("LoginFailureAlarm")

    @property
    async def async_event_login_failure(self) -> str:
        return await self.async_event_handler_config("LoginFailureAlarm")

    @property
    def event_storage_not_exist(self) -> str:
        return self.event_handler_config("StorageNotExist")

    @property
    async def async_event_storage_not_exist(self) -> str:
        return await self.async_event_handler_config("StorageNotExist")

    @property
    def event_storage_access_failure(self) -> str:
        return self.event_handler_config("StorageFailure")

    @property
    async def async_event_storage_access_failure(self) -> str:
        return await self.async_event_handler_config("StorageFailure")

    @property
    def event_storage_low_space(self) -> str:
        return self.event_handler_config("StorageLowSpace")

    @property
    async def async_event_storage_low_space(self) -> str:
        return await self.async_event_handler_config("StorageLowSpace")

    @property
    def event_net_abort(self) -> str:
        return self.event_handler_config("NetAbort")

    @property
    async def async_event_net_abort(self) -> str:
        return await self.async_event_handler_config("NetAbort")

    @property
    def event_ip_conflict(self) -> str:
        return self.event_handler_config("IPConflict")

    @property
    async def async_event_ip_conflict(self) -> str:
        return await self.async_event_handler_config("IPConflict")

    def event_channels_happened(self, eventcode: str) -> List[int]:
        """
        Params:

        VideoMotion: motion detection event
        VideoLoss: video loss detection event
        VideoBlind: video blind detection event
        AlarmLocal: alarm detection event
        StorageNotExist: storage not exist event
        StorageFailure: storage failure event
        StorageLowSpace: storage low space event
        AlarmOutput: alarm output event
        SmartMotionHuman: human detection event
        SmartMotionVehicle: vehicle detection event
        """
        ret = self.command(
            f"eventManager.cgi?action=getEventIndexes&code={eventcode}"
        )
        output = ret.content.decode()
        if "Error" in output:
            return []
        return [int(pretty(channel)) for channel in output.split()]

    async def async_event_channels_happened(self, eventcode: str) -> List[int]:
        ret = await self.async_command(
            f"eventManager.cgi?action=getEventIndexes&code={eventcode}"
        )
        output = ret.content.decode()
        if "Error" in output:
            return []
        return [int(pretty(channel)) for channel in output.split()]

    @property
    def is_motion_detected(self) -> List[int]:
        return self.event_channels_happened("VideoMotion")

    @property
    async def async_is_motion_detected(self) -> List[int]:
        return await self.async_event_channels_happened("VideoMotion")

    @property
    def is_alarm_triggered(self) -> List[int]:
        return self.event_channels_happened("AlarmLocal")

    @property
    async def async_is_alarm_triggered(self) -> List[int]:
        return await self.async_event_channels_happened("AlarmLocal")

    @property
    def is_human_detected(self) -> List[int]:
        return self.event_channels_happened("SmartMotionHuman")

    @property
    async def async_is_human_detected(self) -> List[int]:
        return await self.async_event_channels_happened("SmartMotionHuman")

    @property
    def is_vehicle_detected(self) -> List[int]:
        return self.event_channels_happened("SmartMotionVehicle")

    @property
    async def async_is_vehicle_detected(self) -> List[int]:
        return await self.async_event_channels_happened("SmartMotionVehicle")

    @property
    def event_management(self) -> str:
        ret = self.command("eventManager.cgi?action=getCaps")
        return ret.content.decode()

    @property
    async def async_event_management(self) -> str:
        ret = await self.async_command("eventManager.cgi?action=getCaps")
        return ret.content.decode()

    def event_stream(
        self,
        eventcodes: str,
        *,
        retries: Optional[int] = None,
        timeout_cmd: TimeoutT = None,
    ) -> Iterator[str]:
        """
        Return a stream of event info lines.

        eventcodes: One or more event codes separated by commas with no spaces

        VideoMotion: motion detection event
        VideoLoss: video loss detection event
        VideoBlind: video blind detection event
        AlarmLocal: alarm detection event
        StorageNotExist: storage not exist event
        StorageFailure: storage failure event
        StorageLowSpace: storage low space event
        AlarmOutput: alarm output event
        SmartMotionHuman: human detection event
        SmartMotionVehicle: vehicle detection event
        """
        urllib3_logger = logging.getLogger("urllib3.connectionpool")
        if not any(
            isinstance(x, NoHeaderErrorFilter) for x in urllib3_logger.filters
        ):
            urllib3_logger.addFilter(NoHeaderErrorFilter())

        # If timeout is not specified, then use default, but remove read
        # timeout since there's no telling when, if ever, an event will come.
        if timeout_cmd is None:
            if isinstance(self._timeout_default, tuple):
                timeout_cmd = self._timeout_default[0], None
            else:
                timeout_cmd = self._timeout_default, None

        ret = self.command(
            f"eventManager.cgi?action=attach&codes=[{eventcodes}]",
            retries=retries,
            timeout_cmd=timeout_cmd,
            stream=True,
        )
        if ret.encoding is None:
            ret.encoding = "utf-8"  # type: ignore[unreachable]

        try:
            for line in _event_lines(ret.iter_content(decode_unicode=True)):
                if line.lower().startswith("content-length:"):
                    chunk_size = int(line.split(":")[1])
                    try:
                        yield next(
                            ret.iter_content(
                                chunk_size=chunk_size, decode_unicode=True
                            )
                        )
                    except StopIteration:
                        return
        except (RequestException, HTTPError) as error:
            _LOGGER.debug("%s Error during event streaming: %r", self, error)
            raise CommError(error) from error
        finally:
            ret.close()

    async def async_event_stream(
        self, eventcodes: str, *, timeout_cmd: TimeoutT = None
    ) -> AsyncIterator[str]:
        """Return a stream of event info lines."""
        # If timeout is not specified, then use default, but remove read
        # timeout since there's no telling when, if ever, an event will come.
        if timeout_cmd is None:
            if isinstance(self._timeout_default, tuple):
                timeout_cmd = self._timeout_default[0], None
            else:
                timeout_cmd = self._timeout_default, None

        async with self.async_stream_command(
            f"eventManager.cgi?action=attach&codes=[{eventcodes}]",
            timeout_cmd=timeout_cmd,
        ) as ret:
            it = ret.aiter_text(chunk_size=1)
            async for line in _async_event_lines(it):
                if line.lower().startswith("content-length:"):
                    chunk_size = int(line.split(":")[1])
                    chars = []
                    async for char in it:
                        chars.append(char)
                        if len(chars) == chunk_size:
                            break
                    else:
                        # If we can't get the chunk, then return out
                        return
                    yield "".join(chars)

    def event_actions(
        self,
        eventcodes: str,
        retries: Optional[int] = None,
        timeout_cmd: TimeoutT = None,
    ) -> Iterator[Tuple[str, Dict[str, Any]]]:
        """Return a stream of event (code, payload) tuples."""
        for event_info in self.event_stream(
            eventcodes, retries=retries, timeout_cmd=timeout_cmd
        ):
            payload = self._build_payload(event_info)
            yield payload["Code"], payload

    async def async_event_actions(
        self, eventcodes: str, timeout_cmd: TimeoutT = None
    ) -> AsyncIterator[Tuple[str, Dict[str, Any]]]:
        """Return a stream of event (code, payload) tuples."""
        async for event_info in self.async_event_stream(
            eventcodes, timeout_cmd=timeout_cmd
        ):
            payload = self._build_payload(event_info)
            yield payload["Code"], payload

    def _build_payload(self, event_info: str) -> Dict[str, Any]:
        _LOGGER.debug("%s event info: %r", self, event_info)
        payload = {}
        for key, value in _REG_PARSE_KEY_VALUE.findall(
            event_info.strip().replace("\n", "")
        ):
            if key == "data":
                value = {
                    data_key.replace('"', ""): data_value.replace('"', "")
                    for data_key, data_value in _REG_PARSE_MALFORMED_JSON.findall(
                        value
                    )
                }
            payload[key] = value
        _LOGGER.debug(
            "%s generate new event, code: %s , payload: %s",
            self,
            payload["Code"],
            payload,
        )
        return payload


class NoHeaderErrorFilter(logging.Filter):
    """
    Filter out urllib3 Header Parsing Errors due to a urllib3 bug.

    See https://github.com/urllib3/urllib3/issues/800
    """

    def filter(self, record):
        """Filter out Header Parsing Errors."""
        return "Failed to parse headers" not in record.getMessage()



#!/usr/bin/env python
from setuptools import setup, find_packages


def readme():
    with open("README.rst") as desc:
        return desc.read()


setup(
    name="amcrest",
    version="1.9.4",
    description="Python wrapper implementation for Amcrest cameras.",
    long_description=readme(),
    author="Douglas Schilling Landgraf, Marcelo Moreira de Mello",
    author_email="dougsland@gmail.com, tchello.mello@gmail.com",
    url="http://github.com/tchellomello/python-amcrest",
    license="GPLv2",
    install_requires=[
        "argcomplete",
        "httpx",
        "requests",
        "typing_extensions",
        "urllib3",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={"amcrest": ["py.typed"]},
    test_suite="tests",
    scripts=["cli/amcrest-cli", "tui/amcrest-tui"],
    zip_safe=True,
    keywords="amcrest camera python",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
from typing import Optional

from .http import Http


class UserManagement(Http):
    def info_user(self, username: str) -> str:
        return self._user_manager(f"getUserInfo&name={username}")

    async def async_info_user(self, username: str) -> str:
        return await self._async_user_manager(f"getUserInfo&name={username}")

    @property
    def info_all_users(self) -> str:
        return self._user_manager("getUserInfoAll")

    @property
    async def async_info_all_users(self) -> str:
        return await self._async_user_manager("getUserInfoAll")

    @property
    def info_all_active_users(self) -> str:
        return self._user_manager("getActiveUserInfoAll")

    @property
    async def async_info_all_active_users(self) -> str:
        return await self._async_user_manager("getActiveUserInfoAll")

    def info_group(self, group: str) -> str:
        return self._user_manager(f"getGroupInfo&name={group}")

    async def async_info_group(self, group: str) -> str:
        return await self._async_user_manager(f"getGroupInfo&name={group}")

    @property
    def info_all_groups(self) -> str:
        return self._user_manager("getGroupInfoAll")

    @property
    async def async_info_all_groups(self) -> str:
        return await self._async_user_manager("getGroupInfoAll")

    def delete_user(self, username: str) -> str:
        return self._user_manager(f"deleteUser&name={username}")

    async def async_delete_user(self, username: str) -> str:
        return await self._async_user_manager(f"deleteUser&name={username}")

    def add_user(
        self,
        username: str,
        password: str,
        group: str,
        sharable: bool = True,
        reserved: bool = False,
        memo: Optional[str] = None,
    ) -> str:
        """
        Params:
            username - username for user
            password - password for user
            group - string the range is "admin" and "user". In different group,
                    the user has different authorities.

            sharable - bool, true means allow multi-point login

            reserved - bool, true means this user can't be deleted

            memo - memo to user
        """

        cmd = (
            "addUser"
            f"&user.Name={username}"
            f"&user.Password={password}"
            f"&user.Group={group.lower()}"
            f"&user.Sharable={str(sharable).lower()}"
            f"&user.Reserved={str(reserved).lower()}"
        )

        if memo:
            cmd += f"&user.Memo={memo}"

        return self._user_manager(cmd)

    async def async_add_user(
        self,
        username: str,
        password: str,
        group: str,
        sharable: bool = True,
        reserved: bool = False,
        memo: Optional[str] = None,
    ) -> str:
        cmd = (
            "addUser"
            f"&user.Name={username}"
            f"&user.Password={password}"
            f"&user.Group={group.lower()}"
            f"&user.Sharable={str(sharable).lower()}"
            f"&user.Reserved={str(reserved).lower()}"
        )

        if memo:
            cmd += f"&user.Memo={memo}"

        return await self._async_user_manager(cmd)

    def modify_password(self, username: str, newpwd: str, oldpwd: str) -> str:
        """
        Params:
            username - user name
            newpwd - new password
            oldpwd - old password
        """
        return self._user_manager(
            f"modifyPassword&name={username}&pwd={newpwd}&pwdOld={oldpwd}"
        )

    async def async_modify_password(
        self, username: str, newpwd: str, oldpwd: str
    ) -> str:
        return await self._async_user_manager(
            f"modifyPassword&name={username}&pwd={newpwd}&pwdOld={oldpwd}"
        )

    def modify_user(self, username: str, attribute: str, value: str) -> str:
        """
        Params:
            username - username for user
            attribute - the attribute name that will change:
                        group, sharable, reserved, memo

            value - the new value for attribute
        """

        cmd = f"modifyUser&name={username}"

        if attribute.lower() == "group":
            cmd += f"&user.Group={value.lower()}"

        elif attribute.lower() == "sharable":
            cmd += f"&user.Sharable={value.lower()}"

        elif attribute.lower() == "reserved":
            cmd += f"&user.Reserved={value.lower()}"

        elif attribute == "memo":
            cmd += f"&user.Memo={value.lower()}"

        return self._user_manager(cmd)

    async def async_modify_user(
        self, username: str, attribute: str, value: str
    ) -> str:
        cmd = f"modifyUser&name={username}"

        if attribute.lower() == "group":
            cmd += f"&user.Group={value.lower()}"

        elif attribute.lower() == "sharable":
            cmd += f"&user.Sharable={value.lower()}"

        elif attribute.lower() == "reserved":
            cmd += f"&user.Reserved={value.lower()}"

        elif attribute == "memo":
            cmd += f"&user.Memo={value.lower()}"

        return await self._async_user_manager(cmd)

    def _user_manager(self, action: str) -> str:
        ret = self.command(f"userManager.cgi?action={action}")
        return ret.content.decode()

    async def _async_user_manager(self, action: str) -> str:
        ret = await self.async_command(f"userManager.cgi?action={action}")
        return ret.content.decode()
